#!/usr/bin/python3

# Rev A.01.07 - May 4, 2023
# The DZ Company, LLC
# This software may be distributed freely as long as credit is given to DZKit

import SoapySDR
from SoapySDR import * #SOAPY_SDR_ constants
import threading
import time
import queue
import RPi.GPIO as GPIO
import pyaudio
import numpy as np
from numpy import pi, absolute, arange
from scipy.signal import bilinear, lfilter, lfilter_zi, butter
from scipy import log
import sw_rx_globals as g
#import matplotlib.pyplot as plt
    
class Filter_coeff:
    def __init__(self, b, a):
        self.b = b
        self.a = a
  
def demod():
    # Demodulation
    #t0 = time.perf_counter()
    y = g.last_buf                          # get last buffer
    y = np.append(y, g.x)			        # append current buffer (now total length is 64512 complex elements)
    mode = g.mode
        
    if mode == "FM" or mode == "NBFM":
        g.frq_ofs = 0                       # account for freq shift
        y = np.diff(np.unwrap(np.angle(y))) # demod
        b = g.fmf.b			                # get de-emphasis filter coeffs
        a = g.fmf.a
        zi = lfilter_zi(b, a)*g.ly	   	    # init filter 
        y, zf = lfilter(b, a, y, zi=zi)	    # apply chosen filter
    elif mode == "AM":
        g.frq_ofs = 0                       # account for freq shift
        b = g.bpf6k.b			            # Get 6kHz filter coeffs
        a = g.bpf6k.a
        zi = lfilter_zi(b, a)*g.ly	   	    # init filter 
        y, zf = lfilter(b, a, y, zi=zi)	    # apply chosen filter
    elif mode == "USB":
        g.frq_ofs = 0                       # account for freq shift
        y *= g.exp_jwt_m1r4k                # shift down 1.4kHz
        b = g.bpf2r8k.b			            # get +/-1400 Hz filter coeffs
        a = g.bpf2r8k.a
        zi = lfilter_zi(b, a)*g.ly	   	    # init filter 
        y, zf = lfilter(b, a, y, zi=zi)	    # apply chosen filter
        y *= g.exp_jwt_1r4k                 # shift back up 1400Hz
    elif mode == "LSB":  
        g.frq_ofs = 0                       # account for freq shift
        y *= g.exp_jwt_1r4k                 # shift up 1400Hz
        b = g.bpf2r8k.b			            # get +/-1400 Hz filter coeffs
        a = g.bpf2r8k.a
        zi = lfilter_zi(b, a)*g.ly	   	    # init filter 
        y, zf = lfilter(b, a, y, zi=zi)	    # apply chosen filter
        y *= g.exp_jwt_m1r4k                # shift back down 1400Hz
    elif mode == "CW(USB)":
        g.frq_ofs = 0                       # account for freq shift
        y *= g.exp_jwt_m400                 # shift down 400Hz
        b = g.bpf800.b			            # get +/-400 Hz filter coeffs
        a = g.bpf800.a
        zi = lfilter_zi(b, a)*g.ly	   	    # init filter
        y, zf = lfilter(b, a, y, zi=zi)	    # apply chosen filter
        y = lfilter(b, a, y)        	    # apply chosen filter
        y *= g.exp_jwt_400                  # shift back up 400Hz
    elif mode == "CW(LSB)":  
        g.frq_ofs = 0                       # account for freq shift
        y *= g.exp_jwt_400                  # shift up 400Hz
        b = g.bpf800.b			            # get +/-400 Hz filter coeffs
        a = g.bpf800.a
        zi = lfilter_zi(b, a)*g.ly	   	    # init filter 
        y, zf = lfilter(b, a, y, zi=zi)	    # apply chosen filter
        y *= g.exp_jwt_m400                 # shift back down 400Hz

    g.ly = g.x[-1]                          # get last element of current buf to set initial condition of next filter
 
    if mode == "AM":
        y = np.hypot(y.real, y.imag)	    # demod 
    else:
        y = y.real                          # FM already demodulated, but just need real part of array

    y = y[::g.decimation]	                # take every 8 elements -> 8064 reals
    
    spb = len(y)
    y = y[(spb//4):(3*spb//4)]              # always use middle of the two buffers

    # make sure the gain is such that the value to be output is roughly the same per mode
    if g.mode == "FM":
        maxi = 10000.0 
    elif g.mode == "NBFM":
        maxi = 50000.0
    elif g.mode == "AM":
        maxi = 4.0e6
    elif g.mode == "LSB" or g.mode == "USB":
        maxi = 60.0e6
    else:
        maxi = 150.0e6

    # Apply volume control
    yc = maxi * g.afg / 1024
    y *= yc
    
    # check for overload and limit the peak to 30000
    poverload = np.abs(np.max(y))/30000.0
    noverload = np.abs(np.min(y))/30000.0
    divisor = 1
    if poverload > 1.0:
        divisor = poverload
        #print ("+Overload comp ", poverload, g.iter_nr)

    if noverload > 1.0 and noverload > poverload:
        divisor = noverload
        #print ("-Overload comp ", noverload, g.iter_nr)
        
    y /= divisor

    # Generate max amplitude sine wave for test purposes
    #for i in range (0, 4032):
    #    y[i] = 32767 * np.sin(2*np.pi*0.032*i) * g.afg / 1024
    #print (g.afg)

    #plt.plot(y)
    #plt.show()
    
    g.audio_buf.put(y.astype(np.int16))
    
    #print("Demod done, data added to FIFO. Demod time = ", time.perf_counter()-t0)
    return
    
# Callback to process "framesize" chunks of the audio buffer
def audio_callback(in_data, frame_count, time_info, status):
    g.ab_frame = g.audio_buf.get()
    ret_data = g.ab_frame[0 : frame_count].astype(np.int16).tobytes()
    if g.stopped == True:
        return ret_data, pyaudio.paComplete
    return ret_data, pyaudio.paContinue   

def init_sdr (sdr):
    # SDR setup params
    bandwidth = 200e3
    rx_chan = 0 # only 1 channel on RSP1A

    # init hardware
    sdr.setSampleRate(SOAPY_SDR_RX, rx_chan, g.sample_rate)
    sdr.setFrequency(SOAPY_SDR_RX, rx_chan, g.center_freq)
    sdr.setBandwidth(SOAPY_SDR_RX, rx_chan, bandwidth)
    sdr.setGain(SOAPY_SDR_RX, rx_chan, "RFGR", g.lna)               # gain reduction values go from 0-9 or 0-5; default to 2
    sdr.writeSetting('agc_setpoint', -60)
    if g.agc == "On":
        sdr.setGainMode(SOAPY_SDR_RX, rx_chan, True)                # turn on AGC
    else:
        sdr.setGainMode(SOAPY_SDR_RX, rx_chan, False)               # turn off AGC       
        sdr.setGain(SOAPY_SDR_RX, rx_chan, "IFGR", g.rfg)                   # gain reduction values go from 0-42dB; default to 20

    # Start SDR data stream
    g.rxStream = sdr.setupStream(SOAPY_SDR_RX, SOAPY_SDR_CF32, [rx_chan])
    sdr.activateStream(g.rxStream) 
    return 

def init_filts(show_chart):
    # calculate FM de-emphasis filter params
    g.fmf.b, g.fmf.a = bilinear(1, [75e-6, 1], fs=g.sample_rate)

    #calculate AM Butterworth LPF (+/-3kHz bpf including neg freq)
    wch = 2 * pi * 3000
    b, a = butter(4, wch, btype='low', analog=True)
    g.bpf6k.b, g.bpf6k.a = bilinear(b, a, fs=g.sample_rate)
    
    wch = 2 * pi * 10000
    b, a = butter(4, wch, btype='low', analog=True)
    g.bpf20k.b, g.bpf20k.a = bilinear(b, a, fs=g.sample_rate)
    
    #Shift the filter up 3KHz in frequency by multiplying by exp(j*2*pi*f0*t)
 
    # calculate Butterworth LPF for SSB (+/-2.8kHz bpf incl neg freq)
    wch = 2 * pi * 1400 
    b, a = butter(4, wch, btype='low', analog=True)
    g.bpf2r8k.b, g.bpf2r8k.a = bilinear(b, a, fs=g.sample_rate)
    
    # calculate Butterworth LPF for CW (+/-400Hz bpf incl neg freq)
    wch = 2 * pi * 400 
    b, a = butter(4, wch, btype='low', analog=True)
    g.bpf800.b, g.bpf800.a = bilinear(b, a, fs=g.sample_rate)

    # calc time interval
    Ts = 1.0/g.sample_rate                  # sample period
    t = np.arange(0.0, Ts*(g.num_samps*2), Ts)  # time vector. args are (start, stop, step) (2x buffer length)

    # Shift the filter up in frequency 3kHz by multiplying by exp(j*2*pi*f0*t)    
    f0 = 3.0e3                              # freq shift
    g.exp_jwt_3k = np.exp(2j*np.pi*f0*t)    # this is essentially a complex sine wave
    
    # Shift the filter down in frequency 1.4kHz by multiplying by exp(-j*2*pi*f0*t)
    f0 = -1.4e3
    g.exp_jwt_m1r4k = np.exp(2j*np.pi*f0*t)

    # Shift the filter up in frequency 1.4kHz by multiplying by exp(j*2*pi*f0*t)
    f0 = 1.4e3
    g.exp_jwt_1r4k = np.exp(2j*np.pi*f0*t)
    
    # Shift the filter down in frequency 400Hz by multiplying by exp(-j*2*pi*f0*t)
    f0 = -400
    g.exp_jwt_m400 = np.exp(2j*np.pi*f0*t)

    # Shift the filter up in frequency 400Hz by multiplying by exp(j*2*pi*f0*t)
    f0 = 400
    g.exp_jwt_400 = np.exp(2j*np.pi*f0*t)
    return

def init_audio():
    g.p = pyaudio.PyAudio()
    g.stream = g.p.open(format=pyaudio.paInt16,
                        channels=1,
                        rate=g.drate,
                        output=True,
                        stream_callback=audio_callback,
                        frames_per_buffer=g.framesize)
    g.stream.start_stream()
    return 
    
def read_sdr_thread(sdr):
    while (g.stopped == False):
        read_sdr(sdr)

def read_sdr (sdr):
    g.iter_nr += 1
    for i in range(g.num_rx_calls):
        sr = sdr.readStream(g.rxStream, [g.sample_buffer], len(g.sample_buffer))
        g.x[i*g.max_buffer_size:(i+1)*g.max_buffer_size] = g.sample_buffer

    demod()                                 # run demodulator and filtering algorithm on two buffers
    
    g.last_buf = np.copy(g.x)               # save current buffer for use in next double-wide calc

    if (g.iter_nr == 4): init_audio()       # start audio stream after first two buffers have been put into the fifo

    return

def shut_down_streams(sdr):         
    #shut down the SDR stream
    #print ("Deactivating sdr stream")
    sdr.deactivateStream(g.rxStream)

    #shut down the audio stream
    #print ("Stopping audio stream")
    while g.stream.is_active():
        #print ("Waiting for audio stream to go inactive")
        time.sleep(0.1)
    g.stream.stop_stream()
    #print ("Closing audio stream")
    g.stream.close()
    #print ("Terminating pyAudio")
    g.p.terminate()
    return
